/*
       VDD [==========] GND
SS     PA4 [==========] PA3  SCK
       PA5 [==========] PA2  MISO
DIO1   PA6 [==========] PA1  MOSI
DIO0   PA7 [==========] PA0  (prog)
RESET  PB3 [==========] PB0
UART   PB2 [==========] PB1

xc8-cc decode_v2.c -O2 -mmcu=attiny804 -DF_CPU=5000000 -Wall -Wextra -ffreestanding -funsigned-char
avr-objcopy -Oihex a.out prog.hex

*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>

struct
{
	char *tx_buf;
	int8_t tx_pos;
	int8_t tx_len;

	char *rx_buf;
	int8_t rx_pos;
	int8_t rx_len;
} spi_state;

struct
{
	char *buf;
	int8_t pos;
	int8_t len;
} uart_state;

#define mem_barrier() asm("": : :"memory")

void uart_wait()
{
	mem_barrier();
	while (1) {
		sleep_enable();
		if (!USART0.CTRLA)
			return;
		sleep_cpu();
	}
}

void uart_write(char *buf, int8_t len)
{
	uart_wait();

	uart_state.buf = buf;
	uart_state.pos = 1;
	uart_state.len = len;

	USART0.CTRLA = USART_TXCIE_bm;
	USART0.TXDATAL = buf[0];
}

ISR(USART0_TXC_vect)
{
	USART0.STATUS = USART_TXCIF_bm;

	if (uart_state.pos >= uart_state.len) {
		USART0.CTRLA = 0;
		sleep_disable();
		return;
	}

	USART0.TXDATAL = uart_state.buf[uart_state.pos];
	uart_state.pos++;
}

void spi_wait()
{
	mem_barrier();
	while (1) {
		sleep_enable();
		if (!SPI0.INTCTRL)
			return;
		sleep_cpu();
	}
}

void spi_read(uint8_t addr, char *buf, int8_t len)
{
	spi_wait();

	spi_state.tx_pos = 0;
	spi_state.tx_len = 0;

	spi_state.rx_buf = buf;
	spi_state.rx_pos = -1;
	spi_state.rx_len = len;

	PORTA.OUTCLR = PIN4_bm; // SS low

	SPI0.INTCTRL = SPI_IE_bm;
	SPI0.DATA = addr & 0x7F;
}

void spi_write(uint8_t addr, char *buf, int8_t len)
{
	spi_wait();

	spi_state.tx_buf = buf;
	spi_state.tx_pos = 0;
	spi_state.tx_len = len;

	spi_state.rx_pos = 0;
	spi_state.rx_len = 0;

	PORTA.OUTCLR = PIN4_bm; // SS low

	SPI0.INTCTRL = SPI_IE_bm;
	SPI0.DATA = addr | 0x80;
}

ISR(SPI0_INT_vect)
{
	(void)SPI0.INTFLAGS;

	char rx_byte = SPI0.DATA;
	if (spi_state.rx_pos >= 0 && spi_state.rx_pos < spi_state.rx_len)
		spi_state.rx_buf[spi_state.rx_pos] = rx_byte;
	spi_state.rx_pos++;

	if (spi_state.rx_pos >= spi_state.rx_len && spi_state.tx_pos >= spi_state.tx_len) {
		PORTA.OUTSET = PIN4_bm; // SS high
		SPI0.INTCTRL = 0;
		sleep_disable();
		return;
	}

	char tx_byte = 0;
	if (spi_state.tx_pos >= 0 && spi_state.tx_pos < spi_state.tx_len)
		tx_byte = spi_state.tx_buf[spi_state.tx_pos];
	SPI0.DATA = tx_byte;
	spi_state.tx_pos++;
}

char spi_buf[26];
char uart_buf[22];
char payload_buf[13];

char num_to_hex(uint8_t num)
{
	if (num < 10)
		return '0' + num;
	return 'A' + num - 10;
}

void dump_buf(char *buf, int8_t len)
{
	for (int8_t i = 0; i < len; i++)
	{
		uart_buf[i * 2] = num_to_hex(buf[i] >> 4);
		uart_buf[i * 2 + 1] = num_to_hex(buf[i] & 0xf);
	}

	uart_buf[len * 2] = '\r';
	uart_buf[len * 2 + 1] = '\n';
	uart_write(uart_buf, len * 2 + 2);
}

int8_t decode_payload()
{
	for (int8_t i = 0; i < 13; i++)
		payload_buf[i] = 0;

	int8_t target_byte = 0;
	int8_t target_bit = 6;
	int8_t hold = 1;

	for (int8_t byte = 0; byte < 26; byte++)
	{
		for (int8_t bit = 7; bit >= 0; bit -= 2)
		{
			char val = ((spi_buf[byte] & (1 << bit)) != 0);

			if (val)
				hold = 0;

			if (hold)
				continue;

			payload_buf[target_byte] |= (val << target_bit);

			target_bit--;
			if (target_bit < 0) {
				target_byte++;
				target_bit = 7;
			}
		}
	}

	uint8_t checksum = 0;
	for (int8_t i = 0; i < 9; i++)
		checksum += payload_buf[i];

	return (checksum == payload_buf[9]);
}

#define RegFifo 0x00
#define RegOpMode 0x01
#define RegDataModul 0x02
#define RegBitrateMsb 0x03
#define RegFrfMsb 0x07
#define RegAfcCtrl 0x0B
#define RegLna 0x18
#define RegRxBw 0x19
#define RegAfcBw 0x1A
#define RegOokPeak 0x1B
#define RegOokFix 0x1D
#define RegAfcFei 0x1E
#define RegRssiConfig 0x23
#define RegRssiValue 0x24
#define RegDioMapping1 0x25
#define RegIrqFlags1 0x27
#define RegRssiThres 0x29
#define RegRxTimeout2 0x2B
#define RegSyncConfig 0x2E
#define RegPacketConfig1 0x37
#define RegPacketConfig2 0x3D
#define RegTestLna 0x58
#define RegTestDagc 0x6F

#define RXBW_FIELD(DccFreq, RxBwMant, RxBwExp) (((DccFreq) << 5) | ((RxBwMant) << 3) | (RxBwExp))

uint8_t dynamic_rssi = 170;
int8_t rssi_updated = 0;
int8_t rssi_backoff = 0;

void main()
{
	CCP = CCP_IOREG_gc; // unlock protected registers
	CLKCTRL.MCLKCTRLB = CLKCTRL_PDIV_4X_gc | CLKCTRL_PEN_bm; // 5 MHz

	// configure pins

	// inputs
	PORTA.DIRCLR = PIN6_bm | PIN7_bm; // RFM interrupts
	PORTA.DIRCLR = PIN2_bm; // MISO

	// outputs
	PORTB.DIRSET = PIN2_bm; // UART
	PORTA.DIRSET = PIN1_bm | PIN3_bm | PIN4_bm; // MOSI, SCK, SS

	// MISO pullup
	PORTA.PIN2CTRL = PORT_PULLUPEN_bm;

	// disable input buffers and enable pull-ups on unused pins
	PORTA.PIN5CTRL = PORT_ISC_INPUT_DISABLE_gc | PORT_PULLUPEN_bm;
	PORTB.PIN0CTRL = PORT_ISC_INPUT_DISABLE_gc | PORT_PULLUPEN_bm;
	PORTB.PIN1CTRL = PORT_ISC_INPUT_DISABLE_gc | PORT_PULLUPEN_bm;

	// RFM interrupts
	PORTA.PIN6CTRL = PORT_ISC_RISING_gc;
	PORTA.PIN7CTRL = PORT_ISC_RISING_gc;

	PORTA.OUTSET = PIN4_bm; // SS high

	// configure TCA interrupt
	TCA0.SINGLE.PER = 4880;
	TCA0.SINGLE.INTCTRL = TCA_SINGLE_OVF_bm;
	TCA0.SINGLE.CTRLA = TCA_SINGLE_CLKSEL_DIV1024_gc | TCA_SINGLE_ENABLE_bm;

	// uart
	USART0.CTRLC = USART_CHSIZE_8BIT_gc;
	USART0.BAUD = 174; // 115200
	USART0.CTRLB = USART_TXEN_bm;

	// spi
	SPI0.CTRLB = SPI_SSD_bm;
	SPI0.CTRLA = SPI_MASTER_bm | SPI_PRESC_DIV4_gc | SPI_CLK2X_bm | SPI_ENABLE_bm;

	set_sleep_mode(SLEEP_MODE_IDLE);

	// enable interrupts
	sei();

	// RFM reset
	PORTB.OUTSET = PIN3_bm;
	PORTB.DIRSET = PIN3_bm;
	_delay_us(100);
	PORTB.DIRCLR = PIN3_bm;

	_delay_ms(20);

	spi_write(RegFrfMsb, (char[]){ 0x6C, 0x7A, 0xE1 }, 3); // 433.92 MHz
	spi_write(RegRxBw, (char[]){ RXBW_FIELD(2, 0b00, 1) }, 1); // DCC 4%, 125 kHz

	spi_write(RegLna, (char[]){ 0x00 }, 1); // 50 ohm, auto AGC
	spi_write(RegTestLna, (char[]){ 0x2D }, 1); // reduce noise floor
	spi_write(RegTestDagc, (char[]){ 0x30 }, 1); // enable DAGC

	spi_write(RegBitrateMsb, (char[]){ 0x0F, 0xA0 }, 2); // 8000 bps
	spi_write(RegDataModul, (char[]){ 0x08 }, 1); // packet mode, OOK
	spi_write(RegDioMapping1, (char[]){ 0x70, 0x07 }, 2); // DIO0 - PayloadReady, DIO1 - Timeout, no CLKOUT
	spi_write(RegSyncConfig, (char[]){ 0x98, 0x55, 0x55, 0x55, 0x55 }, 5); // sync word, 4 bytes, 0 errors
	spi_write(RegPacketConfig1, (char[]){ 0x00, 26 }, 2); // fixed length, no decoding, no crc, 26 bytes
	spi_write(RegPacketConfig2, (char[]){ 0xF2 }, 1); // AutoRxRestartOn, no InterPacketRxDelay
	spi_write(RegRxTimeout2, (char[]){ 16 }, 1); // timeout after 16*16 bits

	spi_write(RegRssiThres, (char*)&dynamic_rssi, 1);
	spi_write(RegOpMode, (char[]){ 0x10 }, 1); // enable receiver

	while (1)
	{
		mem_barrier();
		sleep_enable();
		if (PORTA.IN & PIN7_bm) // DIO0, PayloadReady
		{
			spi_read(RegFifo, spi_buf, 26);
			spi_wait();

			if (decode_payload())
				dump_buf(payload_buf, 10);
		}
		else if (PORTA.IN & PIN6_bm) // DIO1, Timeout
		{
			if (rssi_backoff < 3) {
				rssi_backoff++;
				dynamic_rssi--;
				spi_write(RegRssiThres, (char*)&dynamic_rssi, 1);
			}
			spi_write(RegPacketConfig2, (char[]){ 0xF6 }, 1); // RestartRx
			spi_wait();
		}
		else if (rssi_updated)
		{
			spi_write(RegRssiThres, (char*)&dynamic_rssi, 1);
			rssi_updated = 0;
		}
		else
		{
			sleep_cpu();
		}
	}
}

ISR(TCA0_OVF_vect)
{
	TCA0.SINGLE.INTFLAGS = TCA_SINGLE_OVF_bm;

	if (rssi_backoff <= 1) {
		dynamic_rssi++;
		rssi_updated = 1;
		sleep_disable();
	}

	rssi_backoff = 0;
}

ISR(PORTA_PORT_vect)
{
	PORTA.INTFLAGS = 0xFF;
	sleep_disable();
}
